(*  Title:      ISM/ISM.thy
    Id:         $Id: ISM.thy,v 1.34 2003/09/12 16:13:02 dvo Exp $
    Author:     David von Oheimb
    License:    (C) 2002 Siemens AG; any non-commercial use is granted

*)

header {* Interacting State Machines *}

theory ISM
imports Basis
begin

types ('p, 'm) msgss = "'p \<Rightarrow> 'm list"  --{* family of message lists,\\
i.e. i/o port contents or patterns @{typ 'm} indexed by port names @{typ 'p} *}
(*<*)
translations "(p, m) msgss" \<leftharpoondown> (type) "p \<Rightarrow> m list"
(*>*)
types ('c, 'p, 'm, 's, 'l) trans = "(('p,'m) msgss \<times> 's) \<times> ('c \<times> 'l) \<times> (('p,'m) msgss \<times> 's)"
--{* input pattern, local pre-state, command, output pattern, post-state *}

consts 
  "input" :: "('c, 'p, 'm, 's, 'l) trans \<Rightarrow> ('p, 'm) msgss"
 "output" :: "('c, 'p, 'm, 's, 'l) trans \<Rightarrow> ('p, 'm) msgss"
 "cmd"    :: "('c, 'p, 'm, 's, 'l) trans \<Rightarrow> 'c"
 "label"  :: "('c, 'p, 'm, 's, 'l) trans \<Rightarrow> 'l"
 "lstate" :: "('c, 'p, 'm, 's, 'l) trans \<Rightarrow> 's"
 "lstate'":: "('c, 'p, 'm, 's, 'l) trans \<Rightarrow> 's"
defs
   input_def : " input  \<equiv> fst \<circ> fst"
  output_def : "output  \<equiv> fst \<circ> snd \<circ> snd"
  cmd_def    : "cmd     \<equiv> fst \<circ> fst \<circ> snd"
  label_def  : "label   \<equiv> snd \<circ> fst \<circ> snd"
  lstate_def : "lstate  \<equiv> snd \<circ> fst"
  lstate'_def: "lstate' \<equiv> snd \<circ> snd \<circ> snd"
(*<*)
declare input_def [simp] lstate_def  [simp] cmd_def   [simp] 
       output_def [simp] lstate'_def [simp] label_def [simp]
(*>*)

(* The following definitions can be used to prove noninterference properties *)
constdefs
 "Outputs" :: "(('c, 'p, 'm, 's, 'l) trans) set \<Rightarrow> 'l \<Rightarrow> 's \<Rightarrow> (('p, 'm) msgss) set"
 "Outputs transs l s \<equiv> {p'. \<exists> p c s'. ((p,s),(c,l),(p',s')) \<in> transs}"
 "Step" :: "(('c, 'p, 'm, 's, 'l) trans) set \<Rightarrow> 'l \<Rightarrow> ('s \<times> 's) set"
 "Step transs l \<equiv> {(s, s'). \<exists> p p' c. ((p,s),(c,l),(p',s')) \<in> transs}"

record ('c, 'p, 'm, 's, 'l) "ism" = 
  "inputs" :: "'p set"                           --{* input  port names    *}
 "outputs" :: "'p set"                           --{* output port names    *}
 "init"    :: "'s set"                           --{* initial local states *}
 "trans"   :: "(('c, 'p, 'm, 's, 'l) trans) set" --{* transitions          *}

(*<*)
lemma surjective_ism: 
 "a = \<lparr>inputs = inputs a, outputs = outputs a, init = init a, trans = trans a\<rparr>"
by simp

(*>*)
constdefs
  map_ism ::"('s \<Rightarrow> 't) \<Rightarrow> ('c, 'p, 'm, 's, 'l) ism \<Rightarrow> ('c, 'p, 'm, 't, 'l) ism"
 "map_ism f a \<equiv> \<lparr>inputs = inputs a, outputs = outputs a, init = f`(init a), 
                 trans  = (apfst (apsnd f) \<circ> apsnd (apsnd (apsnd f))) ` trans a\<rparr>"
(*<*)
lemma map_ism_conv [simp]: 
"\<And>tr. map_ism f \<lparr>inputs = inp, outputs = out, init = (  ini), trans = tr\<rparr> = 
                 \<lparr>inputs = inp, outputs = out, init = (f`ini), trans = 
                      (\<lambda>((p,s),(c,l),(p',s')). ((p,f s),(c,l),(p',f s'))) ` tr\<rparr>"
by (auto simp add: map_ism_def apsnd_def apfst_def 
         del: image_eqI intro!: image_eqI)

(*>*)

subsection "Renaming"

constdefs 
  ren_trans :: "'p \<Rightarrow> 'p \<Rightarrow> ('c, 'p, 'm, 's, 'l) trans \<Rightarrow> ('c, 'p, 'm, 's, 'l) trans"
 "ren_trans pn pn' \<equiv> \<lambda>((p,s),(c,l),(p',s')). ((lmap_subst pn pn' p ,s ),(c,l),
                                           (lmap_subst pn pn' p',s'))"

  subst_elem :: "'a \<Rightarrow> 'a \<Rightarrow> 'a set \<Rightarrow> 'a set"
 "subst_elem x y A \<equiv> if x \<in> A then A - {x} \<union> {y} else A"

  ren_conn :: "'p \<Rightarrow> 'p \<Rightarrow> ('c, 'p, 'm, 's, 'l) ism \<Rightarrow> ('c, 'p, 'm, 's, 'l) ism"
 "ren_conn pn pn' a \<equiv> \<lparr>inputs  = subst_elem pn pn' ( inputs a),
                       outputs = subst_elem pn pn' (outputs a),
                       init = init a, trans = ren_trans pn pn' ` trans a\<rparr>"


subsection "Well-formedness"

consts 
  ipns :: "('c, 'p, 'm, 's, 'l) ism \<Rightarrow> 'p set" --{* potential  input port names *}
  opns :: "('c, 'p, 'm, 's, 'l) ism \<Rightarrow> 'p set" --{* potential output port names *}
defs
  ipns_def : "ipns a \<equiv> \<Union>lmap_dom`( input`trans a)"
  opns_def : "opns a \<equiv> \<Union>lmap_dom`(output`trans a)"
constdefs 
  wf_ism :: "('c, 'p, 'm, 's, 'l) ism \<Rightarrow> bool"
 "wf_ism a \<equiv> ipns a \<subseteq> inputs a \<and> opns a \<subseteq> outputs a"
--{* input and output may overlap ($\leadsto$ direct feedback) *}
(*<*)
 lemma wf_ismI: 
  "\<lbrakk>\<And>p s c l p' s' pn. \<lbrakk>((p, s), (c, l), (p', s')) \<in> t; p   pn \<noteq> []\<rbrakk> \<Longrightarrow> pn \<in> Is;
    \<And>p s c l p' s' pn. \<lbrakk>((p, s), (c, l), (p', s')) \<in> t; p'  pn \<noteq> []\<rbrakk> \<Longrightarrow> pn \<in> Os\<rbrakk> \<Longrightarrow> 
  wf_ism \<lparr>inputs = Is, outputs = Os, init = i, trans = t\<rparr>"
by (auto simp add: wf_ism_def ipns_def opns_def)

lemma wf_ism_trans_in_inputs: 
  "\<lbrakk>wf_ism a; ((p, s), c_l_p'_s') \<in> trans a\<rbrakk> \<Longrightarrow> lmap_dom p \<subseteq> inputs a"
apply (clarsimp simp add: wf_ism_def ipns_def UN_subset_iff)
apply (erule rev_ballE, fast)
apply (auto simp add: lmap_dom_def)
done
lemma wf_ism_trans_in_outputs: 
  "\<lbrakk>wf_ism a; (p_s, c_l, (p',s')) \<in> trans a\<rbrakk> \<Longrightarrow> lmap_dom p' \<subseteq> outputs a"
apply (clarsimp simp add: wf_ism_def opns_def UN_subset_iff)
apply (erule rev_ballE, fast)
apply (auto simp add: lmap_dom_def)
done

(*>*)
text {*\pagebreak*}
subsection "Runs"

inductive_set
  runs :: "('c, 'p, 'm, 's, 'l) ism \<Rightarrow> ('s list) set"
  for a :: "('c, 'p, 'm, 's, 'l) ism"
where  --{* ignoring commands and global state! *}
--{* for technical reasons, traces are built from right to left *}
 "init": "s \<in> init a \<Longrightarrow> [s] \<in> runs a"
|"step": "\<lbrakk>s#ss \<in> runs a; 
           ((p,s),(c,l),(p',s')) \<in> trans a
          \<rbrakk> \<Longrightarrow> s'#s#ss \<in> runs a"
-- {* Communication with the environment is unrestricted ($\leadsto$ arbitrary input). *}
(*<*)
inductive_cases runs_elim_cases: "ss \<in> runs a"
ML
{* bind_thm ("runs_step2", rearrange_prems [1,0] (thm "runs.step")) *}
(*>*)

types ('c, 'p, 'm, 's, 'l) truns = "((('c, 'p, 'm, 's, 'l) trans) list) set"
inductive_set 
  truns :: "('c, 'p, 'm, 's, 'l) ism  \<Rightarrow> ('c, 'p, 'm, 's, 'l) truns"
  for a :: "('c, 'p, 'm, 's, 'l) ism"
where
--{* for technical reasons, traces are built from right to left *}
 "init": "\<lbrakk>s\<in>init a;((p,s),c,c2) \<in> trans a\<rbrakk> \<Longrightarrow> [((p,s),c,c2)] \<in> truns a"
|"step": "\<lbrakk>(c0,c',(p1,s1))#ts \<in> truns a; 
           ((p1,s1),c,c2) \<in> trans a\<rbrakk> \<Longrightarrow> 
           ((p1,s1),c,c2)#(c0,c',(p1,s1))#ts \<in> truns a"

lemma truns2trans [rule_format]: "ts \<in> truns a \<Longrightarrow> \<forall>t \<in> set ts. t \<in> trans a"
(*<*)
by (erule truns.induct, auto)
(*>*)

constdefs 
   truns2runs :: "(('c, 'p, 'm, 's, 'l) trans) list \<Rightarrow> 's list"
  "truns2runs ts \<equiv> lstate'(hd ts)#map lstate ts"

lemma truns2runs [rule_format]: "ts \<in> truns a \<Longrightarrow> truns2runs ts \<in> runs a"
(*<*)
by (erule truns.induct, 
    auto simp add: truns2runs_def intro: runs.intros)
(*>*)

lemma lstate_in_truns2runs: "t \<in> set ts \<Longrightarrow> lstate t \<in> set (truns2runs ts)"
(*<*)
by (simp add: truns2runs_def)
(*>*)

subsection "Reachability and invariants"

constdefs
  reach :: "('c, 'p, 'm, 's, 'l) ism \<Rightarrow> 's set"
 "reach a \<equiv> \<Union>set`runs a"

  inv :: "('c, 'p, 'm, 's, 'l) ism \<Rightarrow> ('s \<Rightarrow> bool) \<Rightarrow> bool"
 "inv a P \<equiv> \<forall>s \<in> reach a. P s"

(* The following is needed for Noninterference *)
(* first two helpers *)
lemma  runs_ConsD: "\<lbrakk>x # xs \<in> runs ism; s \<in> set xs\<rbrakk> \<Longrightarrow> xs \<in> runs ism"
(*<*)
by (erule runs.cases, auto)
(*>*)

lemma intermediate_states_in_runs: "\<lbrakk>s \<in> set ss; ss \<in> runs ism\<rbrakk> \<Longrightarrow> \<exists>xs. s#xs \<in> runs ism"
(*<*)
by (induct ss, auto dest: runs_ConsD)
(*>*)

theorem Step_preserves_reachability: "\<lbrakk>s \<in> reach ism; (s, s') \<in> Step (trans ism) a\<rbrakk> \<Longrightarrow> s' \<in> reach ism"
(*<*)
apply (auto simp add: Step_def reach_def)
apply (erule runs.cases)
apply  (clarsimp)
apply  (drule runs.init)
apply  (drule (1) runs.step)
apply  (force)
apply (auto)
apply   ((drule (1) runs.step)+, force)+
apply (drule (1) runs_ConsD)
apply (drule (1) intermediate_states_in_runs, clarify)
apply (drule (1) runs.step)
apply (force)
done
(*>*)

lemma invI: 
"\<lbrakk>\<forall>s \<in> init a. P s; 
  \<And>x p s c l p' s'. \<lbrakk>s \<in> reach a; P s; ((p,s),(c,l),(p',s')) \<in> trans a\<rbrakk> \<Longrightarrow> P s'
 \<rbrakk> \<Longrightarrow> inv a P";
(*<*)
apply (unfold inv_def reach_def)
apply safe
apply (subgoal_tac "Ball (set x) P")
apply  blast
apply (erule runs.induct)
apply  simp
apply auto
apply (subgoal_tac "sa \<in> set (sa # ss) \<and> P sa")
apply blast
apply simp
done
(*>*)

lemma invE: "\<lbrakk>inv a P; s \<in> reach a\<rbrakk> \<Longrightarrow> P s"
(*<*)
by (auto simp: inv_def reach_def split_paired_all)
(*>*)

lemma runs_reach: "\<lbrakk>r \<in> runs a; s \<in> set r\<rbrakk> \<Longrightarrow> s \<in> reach a"
(*<*)
by (auto simp add: reach_def)

lemma truns_lstate_reach: 
 "\<lbrakk>ts \<in> truns a; ((p, s), c_p'_s') \<in> set ts\<rbrakk> \<Longrightarrow> s \<in> reach a"
apply (erule truns2runs [THEN runs_reach])
apply (drule lstate_in_truns2runs)
apply simp
done
(*>*)


subsection "Parallel composition"

types ('i, 'a) family = "('i \<Rightarrow> 'a) \<times> 'i set"

consts
  all_inputs  :: "('i, ('c, 'p, 'm, 's, 'l) ism) family \<Rightarrow> 'p set"
  all_outputs :: "('i, ('c, 'p, 'm, 's, 'l) ism) family \<Rightarrow> 'p set"
  closed      :: "('i, ('c, 'p, 'm, 's, 'l) ism) family \<Rightarrow> bool"
text {*\pagebreak*}
defs
  all_inputs_def : "all_inputs  \<equiv> \<lambda>(A,I). \<Union>( inputs \<circ> A)`I"
  all_outputs_def: "all_outputs \<equiv> \<lambda>(A,I). \<Union>(outputs \<circ> A)`I"
  closed_def     : "closed A \<equiv> all_inputs A = all_outputs A"

(*<*)
lemma inputs_in_all_inputs: 
 "i \<in> snd AI \<Longrightarrow> inputs (fst AI i) \<subseteq> all_inputs AI"
by (auto simp add: all_inputs_def)

lemma outputs_in_all_outputs: 
 "i \<in> snd AI \<Longrightarrow> outputs (fst AI i) \<subseteq> all_outputs AI"
by (auto simp add: all_outputs_def ran_def)
(*>*)
constdefs  --{* (static) well-formedness: inputs of peers should not overlap *}
  wf_comp :: "('i, ('c, 'p, 'm, 's, 'l) ism) family \<Rightarrow> bool"
 "wf_comp \<equiv> \<lambda>(A,I). \<forall>i\<in>I. \<forall>j\<in>I. i \<noteq> j \<longrightarrow> inputs (A i) \<inter> inputs (A j) = {}"
  --{* outputs of peers may overlap ($\leadsto$ nondeterministic interleaving) *}
  wf_isms :: "('i, ('c, 'p, 'm, 's, 'l) ism) family \<Rightarrow> bool"
 "wf_isms \<equiv> \<lambda>(A,I). \<forall>i\<in>I. wf_ism (A i)"
(*<*)
lemma wf_ismsI: "\<And>AI. \<lbrakk>\<And>i. i\<in>snd AI \<Longrightarrow> wf_ism (fst AI i)\<rbrakk> \<Longrightarrow> wf_isms AI"
by (auto simp add: wf_isms_def)

lemma wf_ismsD: "\<lbrakk>wf_isms AI; i\<in>snd AI\<rbrakk> \<Longrightarrow> wf_ism (fst AI i)"
by (auto simp add: wf_isms_def)
(*>*)

(* "proper_buffer before after inb outb" formulates that we can go from "before" to "after" by consuming "inb" and appending "outb" *)
constdefs proper_buffer :: "('p, 'm) msgss \<Rightarrow> ('p, 'm) msgss \<Rightarrow> ('p, 'm) msgss \<Rightarrow> ('p, 'm) msgss \<Rightarrow> bool"
"proper_buffer before after inb outb \<equiv> (\<exists>buf. before=(inb .@. buf) \<and> after=(buf .@. outb))"

lemma [simp]: "proper_buffer a a' \<currency> \<currency> = (a = a')"
by (auto simp add: proper_buffer_def)

types ('p, 'm, 's) conf = "('p, 'm) msgss \<times> 's" 
--{* input buffer state and local state *}
(*<*)
translations "(p, m, s) conf" \<leftharpoondown> (type) "(p, m) msgss \<times> s"
(*>*)

inductive_set
   comp_trans :: "('i, ('c, 'p, 'm, 's, 'l) ism) family \<Rightarrow> 
                  (('c, 'p, 'm, ('p, 'm, 'i \<Rightarrow> 's) conf, 'l) trans) set"
 for AI :: "('i, ('c, 'p, 'm, 's, 'l) ism) family"
where --{* including feedback *}
 step: "\<lbrakk>i \<in> snd AI; ((p1, s1), (c, l), (p2, s2)) \<in> trans (fst AI i)\<rbrakk> \<Longrightarrow>
         ((p1[\<lfloor>](-all_outputs AI), (p1[\<lfloor>](all_outputs AI) .@. b, s(i:=s1))), (c, l),
          (p2[\<lfloor>](-all_inputs  AI), (b .@. p2[\<lfloor>](all_inputs  AI), s(i:=s2)))) 
                                                              \<in> comp_trans AI"
(*<*)
inductive_cases comp_trans_elim_cases: "((p1, (b1 , s1)), (c, l), (p2, (b2, s2))) \<in> comp_trans AI"
(*>*)

lemma comp_trans2transD [rule_format]: "((p, b, gs), (c, l), p', b', gs') \<in> comp_trans family
\<Longrightarrow> \<exists>i \<in> snd family. \<exists>p1 s1 p2 s2. ((p1, s1), (c, l), p2, s2) \<in> trans (fst family i)
  \<and> p=p1[\<lfloor>](- all_outputs family) \<and> (gs i = s1) \<and> p'=p2[\<lfloor>](- all_inputs family) \<and> gs'=gs(i := s2)
  \<and> (proper_buffer b b' (p1[\<lfloor>]all_outputs family) (p2[\<lfloor>]all_inputs family))"
(*<*)
by (erule comp_trans.cases, auto simp add: proper_buffer_def)
(*>*)

lemma comp_trans2transD' [rule_format]: "((p, b, gs), (c, l), p', b', gs') \<notin> comp_trans family
\<Longrightarrow> \<forall>i \<in> snd family. (((p1, s1), (c, l), p2, s2) \<notin> trans (fst family i)
  \<or> p\<noteq>p1[\<lfloor>](- all_outputs family) \<or> (gs i \<noteq> s1) \<or> p'\<noteq>p2[\<lfloor>](- all_inputs family) \<or> gs'\<noteq>gs(i := s2)
  \<or> \<not>(proper_buffer b b' (p1[\<lfloor>]all_outputs family) (p2[\<lfloor>]all_inputs family)))"
(*<*)
by (auto simp add: proper_buffer_def, drule_tac b=buf and s=gs in comp_trans.step, auto)
(*>*)

lemma comp_trans2transD'': "(\<forall>gs' p' p b'. ((p, b, gs), (c, l), p', b', gs') \<notin> comp_trans family)
\<Longrightarrow> \<forall>i \<in> snd family. (((p1, s1), (c, l), p2, s2) \<notin> trans (fst family i)
  \<or> (gs i \<noteq> s1) \<or> (\<forall>buf. b\<noteq>(p1[\<lfloor>]all_outputs family .@. buf)))"
(*<*)
by (auto, drule_tac b=buf and s=gs in comp_trans.step, auto)
(*>*)

lemma comp_trans2transI: "\<lbrakk>i \<in> snd family; ((p1, s1), (c, l), p2, s2) \<in> trans (fst family i);
  p = p1[\<lfloor>](- all_outputs family); proper_buffer b b' (p1[\<lfloor>]all_outputs family) (p2[\<lfloor>]all_inputs family);
  gs i = s1; p' = p2[\<lfloor>](- all_inputs family); gs' = gs(i := s2)\<rbrakk>
\<Longrightarrow> ((p, b, gs), (c, l), p', b', gs') \<in> comp_trans family"
(*<*)
by (insert comp_trans.step [where s=gs and ?s1.0="gs i" and ?s2.0="gs' i" and i=i and AI="family"], auto simp add: proper_buffer_def)
(*>*)

lemma all_outputs_compl_empty: "\<lbrakk>wf_isms family; i \<in> snd family; p2 \<in> output ` trans (fst family i)\<rbrakk>
       \<Longrightarrow> p2[\<lfloor>](- all_outputs family) = \<currency>"
apply (auto simp add: all_outputs_def restrict_lmap_def K_Nil_def)
apply (rule ext, auto, unfold wf_isms_def)
apply (subgoal_tac "wf_ism (ac i)", auto)
apply (drule bspec, auto)
apply (drule (1) wf_ism_trans_in_outputs)
apply (auto simp add: lmap_dom_def)
done

lemma all_inputs_compl_empty: "\<lbrakk>wf_isms family; i \<in> snd family; p1 \<in> input ` trans (fst family i)\<rbrakk>
       \<Longrightarrow> p1[\<lfloor>](- all_inputs family) = \<currency>"
(*<*)
apply (auto simp add: all_inputs_def restrict_lmap_def K_Nil_def)
apply (rule ext, auto, unfold wf_isms_def)
apply (subgoal_tac "wf_ism (ac i)", auto)
apply (drule bspec, auto)
apply (drule (1) wf_ism_trans_in_inputs)
apply (auto simp add: lmap_dom_def)
done
(*>*)

lemma closed_implies_empty_output: "\<lbrakk>closed family; wf_isms family\<rbrakk> \<Longrightarrow> \<forall>a s. Outputs (comp_trans family) a s \<in> {{\<currency>},{}}"
(*<*)
apply (auto simp add: Outputs_def closed_def)
apply  (thin_tac "((p, ?s), (c, a), x, ?aa, ?b) \<in> comp_trans family")
apply  (drule "comp_trans.cases")
apply  (auto simp add: restrict_lmap_def lmap_dom_def K_Nil_def)
apply  (rule ext)
apply  (force dest: wf_ismsD wf_ism_trans_in_outputs simp add: lmap_dom_def all_outputs_def)
apply (erule comp_trans.cases, auto)
apply (frule (1) all_outputs_compl_empty, auto)
apply (force simp add: K_Nil_def dest: comp_trans.step)
done
(*>*)

(* disjunct_labels denotes the property, that if two transitions have the same label, they belong to the same ISM; 
 * this is a very practical property for proofs. *)
constdefs disjunct_labels :: "('i, ('c, 'p, 'm, 's, 'l) ism) family \<Rightarrow> bool"
"disjunct_labels \<equiv> \<lambda>(A,I). \<forall>i\<in>I. \<forall>i'\<in>I. \<forall>p1 p1' b b' c l p2 p2' s2 s2'. 
   ((p1, b), (c, l), p2, s2) \<in> trans (A i) \<and> ((p1', b'), (c, l), p2', s2') \<in> trans (A i') \<longrightarrow> i = i'"

lemma disjunct_labelsD: "\<lbrakk>disjunct_labels family; i \<in> snd family; i' \<in> snd family; ((p1, b), (c, l), p2, s2) \<in> trans (fst family i);
        ((p1', b'), (c, l), p2', s2') \<in> trans (fst family i')\<rbrakk> \<Longrightarrow> i = i'"
(*<*)
by (force simp add: disjunct_labels_def)
(*>*)


(* For each ism, we have a set of possible initial states; out of this, construct
 * the set of possible combined initial states.
 *)
inductive_set
  init_states_set :: "('i \<Rightarrow> ('s set)) \<Rightarrow> ('i \<Rightarrow> 's) set"
  for f :: "('i \<Rightarrow> ('s set))"
where
 step: "\<forall>i. (is i) \<in> f i \<Longrightarrow> is \<in> init_states_set f"

constdefs
  ISM_comp :: "('i, ('c, 'p, 'm, 's, 'l) ism) family \<Rightarrow> 
               ('c, 'p, 'm, ('p, 'm, 'i \<Rightarrow> 's) conf, 'l) ism"
 "ISM_comp AI \<equiv> \<lparr>inputs  = all_inputs AI - all_outputs AI,
              \<spacespace>outputs = all_outputs AI - all_inputs AI,
              \<spacespace>init    = (\<lambda>x. (\<currency>,x)) ` (init_states_set (init \<circ> fst AI)),
              \<spacespace>trans   = comp_trans AI\<rparr>"

lemma runs_comp_buffers_in_all_inputs_outputs [rule_format]:  --{* unused *}
"\<lbrakk>r \<in> runs (ISM_comp AI); wf_isms AI\<rbrakk> \<Longrightarrow> 
 \<forall>bs s' cs. r = (bs,s')#cs \<longrightarrow> lmap_dom bs \<subseteq> all_inputs AI \<inter> all_outputs AI"
(*<*)
apply (erule runs.induct)
apply  (simp_all add: ISM_comp_def del: K_Nil_def2)
apply  (clarsimp)
apply (clarsimp elim!: comp_trans_elim_cases)
apply (drule wf_ismsD, simp)
apply (drule wf_ism_trans_in_outputs, simp)
apply (auto simp add: restrict_lmap_def split add: split_if_asm, auto)
apply (simp_all only: lmap_dom_def2 [symmetric])
apply (fast dest: outputs_in_all_outputs)
done
(*>*)

lemma runs_closed_comp_past_induct [rule_format]:
"\<lbrakk>r \<in> runs (ISM_comp AI); closed AI; wf_isms AI;
  \<forall> init_state \<in> init_states_set (init \<circ> fst AI). P \<currency> init_state [];
  \<And>a p1 p2 cs i b c l s s1 s2.
       \<lbrakk>lmap_dom p1 \<subseteq> all_outputs AI; 
        (p1 .@. b, s(i:=s1)) # cs \<in> runs (ISM_comp AI); 
        P (p1 .@. b) (s(i:=s1)) cs; 
        i \<in> snd AI; ((p1, s1), (c, l), (p2, s2)) \<in> trans (fst AI i)
       \<rbrakk> \<Longrightarrow> P (b .@. p2) (s(i:=s2)) ((p1 .@. b, s(i:=s1)) # cs)
 \<rbrakk> \<Longrightarrow> \<forall>bs s cs. r = (bs,s)#cs \<longrightarrow> P bs s cs"
(*<*)
apply (erule runs.induct)
apply (simp_all add: ISM_comp_def closed_def)
apply  (force simp add: K_Nil_def elim:init_states_set.cases)
apply (clarsimp elim!: comp_trans_elim_cases)
apply (drule wf_ismsD, simp)
apply (frule (1) wf_ism_trans_in_inputs)
apply (drule (1) wf_ism_trans_in_outputs)
apply (subgoal_tac "p1[\<lfloor>]all_outputs AI = p1", simp del: restrict_lmap_dom_eq)
apply  (subgoal_tac "p2[\<lfloor>]all_outputs AI = p2",(*all_inputs!*)simp (no_asm_simp))
apply   (simp)
apply  (simp)
apply  (erule subset_trans)
apply  (erule outputs_in_all_outputs)
apply (simp)
apply (erule subset_trans)
apply (erule inputs_in_all_inputs [THEN subset_trans])
apply blast
done
(*>*)

lemma runs_closed_comp_all_imp_all_induct:
"\<lbrakk>cs \<in> runs (ISM_comp AI); closed AI; wf_isms AI;
  \<forall> is \<in> init_states_set (init \<circ> fst AI). P (\<currency>, is) \<longrightarrow> Q (\<currency>, is);
  \<And>a p1 p2 cs i b c l s s1 s2.
     \<lbrakk>lmap_dom p1 \<subseteq> all_outputs AI; 
      (p1 .@. b, s(i:=s1)) # cs \<in> runs (ISM_comp AI); 
      i \<in> snd AI; ((p1, s1), (c, l), (p2, s2)) \<in> trans (fst AI i);
      Ball (set cs) Q; Q (p1 .@. b, s(i:=s1)); 
      Ball (set cs) P; P (p1 .@. b, s(i:=s1)); P (b .@. p2, s(i:=s2))
     \<rbrakk> \<Longrightarrow> Q (b .@. p2, s(i:=s2))
 \<rbrakk> \<Longrightarrow> (\<forall>c\<in>set cs. P c) \<longrightarrow> (\<forall>c\<in>set cs. Q c)"
(*<*)
apply (erule runs.induct)
apply  (simp_all add: ISM_comp_def closed_def)
apply  (force elim: init_states_set.cases)
apply (clarsimp elim!: comp_trans_elim_cases)
apply (drule wf_ismsD, simp)
apply (frule (1) wf_ism_trans_in_inputs)
apply (drule (1) wf_ism_trans_in_outputs)
apply (subgoal_tac "p1[\<lfloor>]all_outputs AI = p1", simp del: restrict_lmap_dom_eq)
apply  (subgoal_tac "p2[\<lfloor>]all_outputs AI = p2",(*all_inputs!*) simp del: restrict_lmap_dom_eq)
apply   (simp)
apply  (simp)
apply  (erule subset_trans)
apply  (erule outputs_in_all_outputs)
apply (simp)
apply (erule subset_trans)
apply (erule inputs_in_all_inputs [THEN subset_trans]) 
apply blast
done

(*>*)
text {*\pagebreak*}
subsection "Composite runs"

types ('i, 'c, 'g) g_trans =  "'i \<Rightarrow> ('g \<times> 'c \<times> 'g) set"
--{* index, input and output patterns, global pre-state, command, post-state *}

inductive_set comp_runs:: "('g \<Rightarrow> ('i, ('c, 'p, 'm, 's, 'l) ism) family) \<Rightarrow> 
     'g \<Rightarrow> ('i, 'c, 'g) g_trans \<Rightarrow> ((('p, 'm, 'g \<times> ('i \<Rightarrow> 's)) conf) list) set"
  for AIs :: "('g \<Rightarrow> ('i, ('c, 'p, 'm, 's, 'l) ism) family)" 
  and g0 :: "'g" and g_trans :: "('i, 'c, 'g) g_trans"
where
--{* including feedback *}
--{* for technical reasons, traces are built from right to left *}
 "init": "s \<in> init_states_set (init \<circ> fst (AIs g0)) \<Longrightarrow> [(\<currency>, (g0, s))] \<in> comp_runs AIs g0 g_trans"
|"step": "\<lbrakk>(p1 .@. b, (g1, s(i:=s1)))#cs \<in> comp_runs AIs g0 g_trans;
           i \<in> snd (AIs g1); a = fst (AIs g1) i;
           ((p1, s1), (c, l), (p2, s2)) \<in> trans a;
           lmap_dom p1 \<subseteq> inputs a \<inter> all_outputs (AIs g1); 
           lmap_dom p2 \<subseteq> outputs a \<inter> all_inputs (AIs g1); 
           (g1, c, g2) \<in> g_trans i
           \<rbrakk> \<Longrightarrow> 
            (b .@. p2, (g2, s(i:=s2)))#
            (p1.@. b , (g1, s(i:=s1)))#cs \<in> comp_runs AIs g0 g_trans"
-- {* Changes to global state take effect after output has been performed.  *}
-- {* Here the environment is assumed to be part of the ISM system;\\\hspace*{2.5ex}
      communication with the environment may be modeled with suitable ISMs. *}
(*<*)
inductive_cases comp_runs_elim_cases: "cs \<in> comp_runs AIs g0 g_trans"
(*>*)

lemma comp_runs_global_induct: "cs \<in> comp_runs AIs ginit g_trans \<Longrightarrow> 
 P (AIs ginit) \<Longrightarrow> 
 (\<And>g1 c g2 i. \<lbrakk>(g1, c, g2) \<in> g_trans i; i \<in> snd (AIs g1); P (AIs g1)\<rbrakk> \<Longrightarrow>
                                                           P (AIs g2)      ) \<Longrightarrow>
 \<forall>(b,g,s) \<in> set cs. P (AIs g)"
(*<*)
by (erule comp_runs.induct, auto)
(*>*)

(*
lemma past_in_comp_runs [rule_format]: 
 "cs' \<in> comp_runs AIs g0 g_trans \<Longrightarrow> 
 \<forall>c. c \<in> set cs' \<longrightarrow> (\<exists>cs. cs' >= c#cs \<and> c#cs \<in> comp_runs AIs g0 g_trans)"
apply (erule comp_runs.induct)
apply  safe
apply  (force intro: comp_runs.init)
apply (drule set_ConsD, safe)
apply  (blast intro: comp_runs.step)
apply (fast intro: postfix_ConsI)
done
*)

lemma comp_runs_past_induct [rule_format]:
"\<lbrakk>r \<in> comp_runs AIs g0 g_trans; 
  \<forall> s \<in> init_states_set (init \<circ> fst (AIs g0)). P \<currency> g0 s [];
  \<And>a p1 p2 cs i b c l s g1 s1 g2 s2.
       \<lbrakk>(p1 .@. b, (g1, s(i:=s1))) # cs \<in> comp_runs AIs g0 g_trans; 
        P (p1 .@. b) g1 (s(i:=s1)) cs; 
        i \<in> snd (AIs g1);
        ((p1, s1), (c, l), (p2, s2)) \<in> trans (fst (AIs g1) i);
        (g1, c, g2) \<in> g_trans i
       \<rbrakk> \<Longrightarrow> P (b .@. p2) g2 (s(i:=s2)) ((p1 .@. b, (g1, s(i:=s1))) # cs)
 \<rbrakk> \<Longrightarrow> \<forall>bs g s cs. r = (bs,(g,s))#cs \<longrightarrow> P bs g s cs"
(*<*)
apply (erule comp_runs.induct)
apply  (simp_all del: K_Nil_def2)
done
(*>*)

lemma comp_runs_all_imp_all_induct:
"\<lbrakk>cs \<in> comp_runs AIs g0 g_trans; 
  \<forall> s \<in> init_states_set (init \<circ> fst (AIs g0)). P (\<currency>, g0, s) \<longrightarrow> Q (\<currency>, g0, s);
  \<And>a p1 p2 cs i b c l s g1 s1 g2 s2.
     \<lbrakk>(p1 .@. b, g1, s(i:=s1)) # cs \<in> comp_runs AIs g0 g_trans; 
      i \<in> snd (AIs g1); ((p1, s1), (c, l), (p2, s2)) \<in> trans (fst (AIs g1) i);
      (g1, c, g2) \<in> g_trans i; 
      Ball (set cs) Q; Q (p1 .@. b, g1, s(i:=s1)); 
      Ball (set cs) P; P (p1 .@. b, g1, s(i:=s1)); 
      P (b .@. p2, g2, s(i:=s2))
     \<rbrakk> \<Longrightarrow> Q (b .@. p2, g2, s(i:=s2))
 \<rbrakk> \<Longrightarrow> (\<forall>c\<in>set cs. P c) \<longrightarrow> (\<forall>c\<in>set cs. Q c)"
(*<*)
apply (erule comp_runs.induct)
apply  (simp_all)
done
(*>*)

(*<*)
lemma comp_runs_g_trans_idD:
"(p, g, s) # cs \<in> comp_runs AIs g0 (\<lambda>i. {u. \<exists>g c. u = (g, c, g)}) \<Longrightarrow> g = g0"
by (drule_tac P = "\<lambda>bs g s cs. g = g0" in comp_runs_past_induct, auto)
(*>*)
theorem closed_comp_runs_comp: 
"\<lbrakk>wf_isms (AIs g); closed (AIs g)\<rbrakk> \<Longrightarrow> runs (ISM_comp (AIs g)) = 
  map (\<lambda>(b,g,s). (b,s)) ` comp_runs AIs g (\<lambda>i. {(g,c,g')| g c g'. g = g'})"
(*<*)
apply (auto simp: ISM_comp_def)
apply  (erule runs.induct)
apply   simp_all
apply   (clarsimp elim: init_states_set.cases)
apply   (rule image_eqI)
prefer 2
apply    (rule_tac [1] comp_runs.init)
apply    (simp add: o_def)
apply   (simp add: K_Nil_def)
apply  (tactic {* split_all_tac 1*}, rename_tac b s b' s' cs)
apply  (erule comp_trans_elim_cases)
apply  (clarsimp)
apply  (drule wf_ismsD, simp)
apply  (frule (1) wf_ism_trans_in_inputs)
apply  (drule (1) wf_ism_trans_in_outputs)
apply  (frule comp_runs_g_trans_idD, clarify)
apply  (rule image_eqI)
prefer 2
apply   (erule (1) comp_runs.step, rule refl)
prefer 5 apply (force)
apply     (subgoal_tac "p1[\<lfloor>]all_outputs (AIs g) = p1")
apply      (subgoal_tac "p2[\<lfloor>]all_inputs (AIs g) = p2")
apply       (simp (no_asm_simp))
apply      (simp)
apply      (erule subset_trans)
apply      (simp add: closed_def)
apply      (erule outputs_in_all_outputs)
apply     (simp add: closed_def)
apply     (drule sym)
apply     (simp, erule subset_trans)
apply     (erule inputs_in_all_inputs)
apply    force
apply   force
apply  fast
apply (erule_tac V = "closed (AIs g)" in thin_rl)
apply (erule comp_runs.induct, simp_all)
apply  (rule ssubst [OF arg_cong], rule_tac [2] runs.init) back
apply   (force simp add: K_Nil_def o_def elim:init_states_set.cases)
apply  (force simp add: K_Nil_def o_def elim:init_states_set.cases)
apply (rule runs.step, assumption)
apply (frule comp_runs_g_trans_idD)
apply clarsimp
apply (subgoal_tac "p1[\<lfloor>]all_outputs (AIs g) = p1")
apply  (rotate_tac -1, erule subst)
apply  (subgoal_tac "p2[\<lfloor>]all_inputs (AIs g) = p2")
apply   (rotate_tac -1, erule subst)
apply   (drule (2) comp_trans.step)
apply  auto
done
(*>*)

end
