(*  Title:      ISM_package.thy
    ID:         $Id: ISM_package.thy,v 1.13 2003/09/12 16:13:02 dvo Exp $
    Authors:    Sebastian Nanz, TU Muenchen, David von Oheimb, Siemens AG and Thomas Bleher, LMU Muenchen

Definition of Interacting State Machines (ISMs).

TODO: when no "transitions" subsection is present, yields error: Unknown constant: "<thy_name>.<ism_name>.transs"
*)

(*<*)
theory ISM_package
(*TODO 2009: imports dAmbISM*)
imports ISM

begin

ML {*
datatype ('a,'b) sum = Inl of 'a | Inr of 'b;

fun option_map f  NONE    = NONE
|   option_map f (SOME x) = SOME (f x)

type ind_def = {defs: thm list, mono: thm, elims: thm list, intrs: thm list,
            induct: thm, unfold: thm, mk_cases: string -> thm, raw_induct: thm};


(* old-style read functions *)

fun read_ctyp thy s =
  Thm.ctyp_of thy (Syntax.read_typ_global thy s);

fun read_cterm thy (s, T) =
  Thm.cterm_of thy (OldGoals.simple_read_term thy T s);


(* Diese Signatur ist veraltet und wird auch nirgens genutzt. Bei Bedarf die Typen der 
 * beiden Funktionen anpassen und die Kommentarzeichen entfernen.
 *)
(*
signature ISM_PACKAGE =
sig
  type 'a rule
  val add_ism: string -> string -> (string * string) list ->
               string -> string -> string -> string ->
               (string * string option) option ->
               string option ->
               (string * (string option * string)) option ->
               (string * (string option * string)) option ->
               string rule list -> theory ->
               theory * ind_def list * thm list
  val add_ism_i: string -> (string * typ) list ->
               typ -> term -> term -> typ ->
               (typ * term option) option ->
               typ option ->
               (typ * (term option * string)) option ->
               (typ * (term option * term  )) option ->
               term rule list -> theory ->
               theory * ind_def list * thm list
end;
*)

structure ISM_package (*: ISM_PACKAGE*) =
struct

(*** utilities ***)

(** messages and errors **)

val quiet_mode = ref false;
fun message s = if !quiet_mode then () else writeln s;

exception ism_exception;
fun ism_error msg = (raise ism_exception) handle ism_exception => error msg;

fun no_state_error name = "ISM " ^ quote name ^
                         " has neither a control nor a data state declaration.";
fun no_ctrl_st_error name = "ISM has a control state transition in rule " ^
                            quote name ^ " but no control state declaration.";
fun no_ctrl_trans_error name =
        "ISM has control state but not control state transitions in rule " ^
        quote name ^ ".";
fun no_data_st_error name = "ISM has postconditions in transition " ^
                            quote name ^ " but no data state declaration.";
fun field_type_error field_name = "Unable to determine field type for field " ^
                                  quote field_name ^ ".";
fun fetch_constr_error l_st_name g_st_name = "Unable to determine a datatype " ^
                   quote g_st_name ^ "\n that has a single argument of type " ^
                   quote l_st_name ^ ".";

(** constructor of global type **)

fun fetch_constr'' l_st_Typ constr_opt [] = constr_opt
  | fetch_constr'' l_st_Typ constr_opt (constr::rest) =
    let
      val (constr_Term, constr_Typ) = constr;
      val l_st_DTyp = DatatypeAux.dtyp_of_typ [] l_st_Typ;
      val constr_opt' = if constr_Typ = [l_st_DTyp] then SOME constr_Term
                                                    else constr_opt;
    in
      fetch_constr'' l_st_Typ constr_opt' rest
    end;

fun fetch_constr' l_st_Typ constr_opt [] = constr_opt
  | fetch_constr' l_st_Typ constr_opt (descr::rest) =
    let
      val (_, (_, _, constr)) = descr;
      val constr_opt' = fetch_constr'' l_st_Typ constr_opt constr;
    in
      fetch_constr' l_st_Typ constr_opt' rest
    end;

fun fetch_constr l_st_Typ g_st_Typ thy =
  let
    val sign = thy;
    val fetch_constr_err = fetch_constr_error (Syntax.string_of_typ_global sign l_st_Typ)
      (Syntax.string_of_typ_global sign g_st_Typ);

    (* determine type name of global state datatype *)
    val g_st_name = case g_st_Typ of Type (g_st_name, _)
                         => Sign.intern_type sign g_st_name
                     | _ => let val _ = ism_error fetch_constr_err in "" end;

    (* symbol table lookup *)
    val symtab_opt = Symtab.lookup (DatatypePackage.get_datatypes thy) g_st_name;
    val descr_opt = option_map (#descr) symtab_opt;
    val descr = case descr_opt of SOME descr => descr
      | NONE => let val _ = ism_error fetch_constr_err in [] end;
    val constr_opt = fetch_constr' l_st_Typ NONE descr;
    val constr = case constr_opt of NONE =>
      let val _ = ism_error fetch_constr_err in "" end
      | SOME(constr) => constr;
  in
    constr
  end;

fun mk_global_st_term l_st_Typ g_st_Typ_opt thy t =
  case g_st_Typ_opt of NONE           => t
                     | SOME(g_st_Typ) =>
    Const (fetch_constr l_st_Typ g_st_Typ thy, l_st_Typ --> g_st_Typ) $ t

(* end; *)

(** helpers **)

fun combine [] [] = []
  | combine (x::xs) [] = []
  | combine [] (y::ys) = []
  | combine (x::xs) (y::ys) = (x, y) :: combine xs ys;

local
  fun esc_underscores [] = []
  |   esc_underscores (x::xs) = if (x = "_") then "'"::x::esc_underscores xs
                                             else      x::esc_underscores xs;
in
  fun escape_underscores name = implode (esc_underscores (explode name))
end;


(*** type handling ***)

(* 'a rule is the type of an ism transition *)
type 'a rule = ((bstring * Args.src list) * (('a * 'a) option * ('a option * ('a option *
               ('a list * ((('a * bool) * 'a) list * ((('a * bool) * 'a) list *
               ('a option * (((string * 'a) list, 'a) sum)))))))));
(* == ((name, args), (cs_exprs, (label_opt, (dom_opt, (pres, (inps, (outps, (cmd_opt, post_v)))))))) *)
val dummyT = Type ("_dummy", []);

fun mk_Typ prep_typ sign type_constr = typ_of (prep_typ sign type_constr);
fun mk_listT T = Type ("List.list", [T]);


(*** term handling ***)

val dummy_Term = Const("_dummy", dummyT);

fun mk_Term prep_term sign t = term_of (prep_term sign t);

(* K_Nil is the empty message family (i.e. no inputs/outputs); also known as \<currency> (see ISM.thy) *)
fun mk_k_nil_Term sign pt_Typ msg_Typ =
  let
    val msgss_Typ = mk_listT msg_Typ;
    val k_nil_Typ = pt_Typ --> msgss_Typ;
    val k_nil_Term = Const (@{const_name K_Nil}, k_nil_Typ);
  in
    (k_nil_Term, k_nil_Typ)
  end;

(* create pairs *)
fun mk_pair_Const sign a_Typ b_Typ =
  let
    val pair_Const_Typ = a_Typ --> b_Typ --> HOLogic.mk_prodT (a_Typ, b_Typ);
    val pair_Const = Const(@{const_name Pair}, pair_Const_Typ);
  in
    (pair_Const, pair_Const_Typ)
  end;

fun mk_pair_Term_Typ sign (a_Term, a_Typ) (b_Term, b_Typ) =
  let
    val (pair_Const, _) = mk_pair_Const sign a_Typ b_Typ;
    val pair_Term = pair_Const $ a_Term $ b_Term;
    val pair_Typ = HOLogic.mk_prodT(a_Typ, b_Typ);
  in
    (pair_Term, pair_Typ)
  end;

fun mk_state_opt prep_type prep_term sign = option_map
   (fn (s_type, (s_term, s_Var)) =>
    let
      val s_Typ  = mk_Typ  prep_type sign s_type;
      val s_Term = case s_term of NONE   => NONE
                                | SOME t => SOME (mk_Term prep_term sign (t, s_Typ));
    in
      (s_Term, s_Typ, s_Var)
    end);



(*** transitions ***)

(** prepare inputs and outputs **)

fun mk_msgss_pat sign pt_Typ msg_Typ =
  let
    val msg_list_Typ = mk_listT msg_Typ;
    val msgss_Typ = pt_Typ --> msg_list_Typ;
    fun fun_upd_Typ T = msgss_Typ --> T --> msg_list_Typ --> msgss_Typ;
    val fun_upd_Term   = Const (@{const_name fun_upd},
                                fun_upd_Typ pt_Typ);
    val fun_upd_s_Term = Const (@{const_name fun_upd_s},
                                fun_upd_Typ (HOLogic.mk_setT pt_Typ));
   in
      Library.foldl (fn (f, ((pat,multi), msgs)) =>
          (if multi then fun_upd_s_Term else fun_upd_Term) $ f $ pat $ msgs)
    end;


(** update record fields (for postconditions) **)
(* TODO: I think we don't handle record extensions correctly here\<dots> This code may break
 * if  the user passes in a complex extended record.
 *)
fun find_field_Typ sign field_name data_st_Typ =
  let
    val record_fields = fst(RecordPackage.get_recT_fields sign data_st_Typ);
    val found_types = filter (fn x => Long_Name.base_name(fst(x)) = field_name) record_fields;
  in
    case found_types of
      []  => let val _ = ism_error (field_type_error field_name) in dummyT end
      | _ => snd(hd(found_types))
  end;

(* Typ: Context.theory -> Term.typ -> Term.term -> (string * Term.term) list -> Term.term *)
fun mk_field_upd thy data_st_Typ r [] = r
  | mk_field_upd thy data_st_Typ r (post::rest) =
    let
      val (field_name, field_Term) = post;
      val field_Typ = find_field_Typ thy field_name data_st_Typ;
      val K_field_Term = Sign.cert_term thy (absdummy (field_Typ, field_Term));
      val field = Const(Sign.intern_const thy (field_name ^ "_update"),
                        (field_Typ --> field_Typ) --> data_st_Typ --> data_st_Typ)
                  $ K_field_Term $ r;
    in
      mk_field_upd thy data_st_Typ field rest
    end;


(** add premises terms to conclusion term **)

fun mk_prems sign concl [] = concl
  | mk_prems sign concl (pre_Term::rest) =
    let
      val impl_Term = Const("==>", propT --> propT --> propT);
      val concl' = impl_Term $ (HOLogic.mk_Trueprop pre_Term) $ concl;
    in
      mk_prems sign concl' rest
    end;

(** transform transition rule (for simultaneous type check) **)

(* This function transforms a "string rule" as returned by the parser into a
 * "term rule" which is then used by the other functions.
 *)
fun transform_rule prep_rule sign types (rule:''a rule) =
  let
    val (cmd_Typ_default_opt, pt_Typ, msg_Typ, st_Typ, ctrl_st_Typ_Term_opt, data_st_Typ_Term_opt,
      l_st_Typ, g_st_Typ_opt, labels_Typ_default_opt, domain_Typ_default_opt) = types;
    val ((name, args), (cs_exprs, (label_opt, (dom_opt, (pres, (inps, (outps, (cmd_opt, post_v)))))))) =
        rule;
    val msg_list_Typ = mk_listT msg_Typ;
    val ctrl_st_Typ = case ctrl_st_Typ_Term_opt of
                        NONE => (case cs_exprs of NONE => dummyT
                                  | SOME _ => ism_error (no_ctrl_st_error name))
                       | SOME (_, typ, _) => (case cs_exprs of NONE =>
                          ism_error (no_ctrl_trans_error name) | SOME _ => typ);

    (* transform rule into Term_Typ list (for prep_rule) *)
    val l = case data_st_Typ_Term_opt of NONE => []
      | SOME(_, data_st_Typ, data_st_var) => [(data_st_var, data_st_Typ)];
    val l = l @ (case cs_exprs of NONE => [] | SOME (cs_expr,cs_expr') =>
                 [(cs_expr, ctrl_st_Typ), (cs_expr', ctrl_st_Typ)]);
    val l = l @ (map (fn p => (p, HOLogic.boolT)) pres);
    val trans_forw_pats = Library.flat o map (fn ((pat, multi), msgs) =>
                       [(pat, if multi then HOLogic.mk_setT pt_Typ else pt_Typ),
                        (msgs, msg_list_Typ)]);
    val l = l @ trans_forw_pats inps;
    val l = l @ trans_forw_pats outps;
    val l = l @ (case cmd_opt of NONE => []
    | SOME cmd => (case cmd_Typ_default_opt of
                     NONE => ism_error ("ISM has no commands " ^
                      "but subsection \"cmd\" is present in rule " ^quote name)
                   | SOME (cmd_Typ, _) => [(cmd, cmd_Typ)]));
    val l = l @ (case label_opt of NONE => []
    | SOME label => (case labels_Typ_default_opt of
                     NONE => ism_error ("ISM has no labels " ^
                      "but subsection \"label\" is present in rule " ^quote name)
                   | SOME (label_Typ, _) => [(label, label_Typ)]));
    val l = l @ (case dom_opt of NONE => []
    | SOME dom => (case domain_Typ_default_opt of
                     NONE => ism_error ("ISM has no domains " ^
                      "but subsection \"dom\" is present in rule " ^quote name)
                   | SOME (dom_Typ, _) => [(dom, dom_Typ)]));

    val l = case data_st_Typ_Term_opt of NONE => (if (post_v <> Inl []) then
                                   ism_error (no_data_st_error name) else l)
                              | SOME (_, data_st_Typ, _) => l @ (case post_v of
          Inl ps => (map (fn (f,v) => (v,find_field_Typ sign f data_st_Typ))) ps
        | Inr sval => [(sval, data_st_Typ)]);

    (* transformation *)
    val l = map term_of (prep_rule sign l);

    (* transform Term_Typ list back into rule *)
    val (data_st_var_opt,l) = case data_st_Typ_Term_opt of NONE => (NONE,l)
                              | _ => (SOME (hd l), Library.drop (1, l));
    val (ctrl_st_exprs,l) = (case cs_exprs of NONE => (NONE,l)
                             | SOME _ => (SOME (hd l, hd (tl l)), Library.drop(2, l)));
    val pres' = Library.take (length pres, l);
    val l = Library.drop (length pres, l);
    fun trans_back_pats l [] = []
    |   trans_back_pats l (((pat,multi),_)::rest) =
        ((hd l, multi), hd (tl l))::trans_back_pats (tl (tl l)) rest;
    val inps' = trans_back_pats l inps;
    val l = Library.drop (2* length inps, l);
    val outps' = trans_back_pats l outps;
    val l = Library.drop (2* length outps, l);
    val (cmd_opt',l) = if cmd_opt = NONE then (NONE,l)
                       else (SOME (hd l), Library.drop (1, l));
    val (label_opt',l) = if label_opt = NONE then (NONE,l)
                         else (SOME (hd l), Library.drop (1, l));
    val (dom_opt',l) = if dom_opt = NONE then (NONE,l)
                       else (SOME (hd l), Library.drop (1, l));
    val post_v' = case post_v of
          Inl posts => Inl (combine ((map (fn (f, v) => f)) posts) l)
        | Inr sval =>  Inr (hd l);
    val rule' = ((name, args),
                (ctrl_st_exprs, (label_opt', (dom_opt', (pres', (inps', (outps', (cmd_opt', post_v'))))))));
  in
    (data_st_var_opt, (rule':term rule))
  end;

(** make introduction rules **)
(* the transs relation is defined inductively in this function *)
fun mk_intros_trans prep_rule thy types param_transs_Term rule =
    let
      val sign = thy;
      (* expand types *)
      val (cmd_Typ_default_opt, pt_Typ, msg_Typ, st_Typ, ctrl_st_Typ_Term_opt, data_st_Typ_Term_opt,
        l_st_Typ, g_st_Typ_opt, labels_Typ_default_opt, _) = types;

      val (data_st_var_opt, rule') =
               transform_rule prep_rule sign types rule;

      (* expand transition *)
      val ((name, args), (ctrl_st_exprs, conds)) = rule';
      val (label_opt,( _, (pres, (inps, (outps, (cmd_opt, post_v)))))) = conds;

      val (k_nil_Term, _) = mk_k_nil_Term sign pt_Typ msg_Typ;
      val msgss_Typ = pt_Typ --> mk_listT msg_Typ;

      (* prepare inputs and outputs; this creates terms of the form "\<currency>(SomePort := [Msg])" *)
      val is_Term = mk_msgss_pat sign pt_Typ msg_Typ (k_nil_Term, inps);
      val os_Term = mk_msgss_pat sign pt_Typ msg_Typ (k_nil_Term, outps);

      (* prepare states *)
      fun mk_state proj post_v = mk_global_st_term l_st_Typ g_st_Typ_opt thy
        (case data_st_Typ_Term_opt of
          NONE => proj (the ctrl_st_exprs)
        | SOME(_, data_st_Typ, _) =>
          let
            val data_st_Term = case post_v of
                    Inl posts => mk_field_upd sign data_st_Typ
                                              (the data_st_var_opt) posts
                  | Inr sval => sval;
          in case ctrl_st_Typ_Term_opt of
               NONE => data_st_Term
             | SOME(_, ctrl_st_Typ, _) =>
                 fst (mk_pair_Const sign ctrl_st_Typ data_st_Typ)
                 $ proj (the ctrl_st_exprs) $ data_st_Term
	  end);
      val s_Term  = mk_state fst (Inl []);
      val s'_Term = mk_state snd post_v;

      (* construct transition *)
      val cmd_Typ = case cmd_Typ_default_opt of SOME (t,_) => t | NONE =>
                                                                  HOLogic.unitT;
      val cmd     = case cmd_opt of SOME t => t | NONE =>
                   (case cmd_Typ_default_opt of
                 NONE => HOLogic.unit
               | SOME (_, SOME default_cmd) => default_cmd
               | _ => ism_error ("ISM has commands without default command " ^
                    "but subsection \"cmd\" is missing in rule " ^ quote name));

      val label_Typ = case labels_Typ_default_opt of SOME (t,_) => t
                                                   | NONE       => HOLogic.unitT;
      val label = case label_opt of SOME t => t | NONE =>
                   (case labels_Typ_default_opt of
                      NONE => HOLogic.unit
                    | SOME (_, SOME default_label) => default_label
                    | _ => ism_error ("ISM has labels without default labels " ^
                           "but subsection \"label\" is missing in rule " ^ quote name));

      (* create one trans tuple *)
      val (is_s_Term, is_s_Typ) =
           mk_pair_Term_Typ sign (is_Term, msgss_Typ) (s_Term, st_Typ);
      val (os_s'_Term, os_s'_Typ) =
           mk_pair_Term_Typ sign (os_Term, msgss_Typ) (s'_Term, st_Typ);
      val (cd_Term, cd_Typ) =
           mk_pair_Term_Typ sign (cmd, cmd_Typ) (label, label_Typ);
      val (trans_Term, trans_Typ) = mk_pair_Term_Typ sign (is_s_Term, is_s_Typ)
        (mk_pair_Term_Typ sign (cd_Term, cd_Typ) (os_s'_Term, os_s'_Typ));

      (* construct conclusion *)
      val mem_const = Const(@{const_name "op :"},
                    trans_Typ --> HOLogic.mk_setT trans_Typ --> HOLogic.boolT);
      val concl_Term = HOLogic.mk_Trueprop
                       (mem_const $ trans_Term $ param_transs_Term);

      (* construct rule with premissions *)
      val rule_Term = mk_prems sign concl_Term (rev pres);

      (* introduction rule *)
      val args' = map (Attrib.intern_src thy) args;
      val intr = ((Binding.name name, args'), rule_Term);
    in
      intr
    end;

(** add transition set **)
(* If you use something newer than Isabelle2005, use OldInductivePackage instead *)
local
  open InductiveSetPackage
in
fun add_transs prep_rule types param_transs_Term [] thy = thy
|   add_transs prep_rule types param_transs_Term transs thy =
    let
      val intros = map (mk_intros_trans prep_rule thy types param_transs_Term) transs;
      val cintros = map (apsnd (Sign.cert_term thy)) intros;
      val (Free (x, T), ps) = strip_comb (Sign.cert_term thy param_transs_Term);
      val thy' = thy
        |> TheoryTarget.init NONE
        |> add_inductive_i
          {quiet_mode = false,
           verbose = true,
           kind = "",
           alt_name = Binding.empty,
           coind = false,
           no_elim = false,
           no_ind = false,
           skip_mono = false,
           fork_mono = false}
          [((Binding.name x, T), NoSyn)] (map dest_Free ps) cintros []
        |> snd
        |> LocalTheory.exit
        |> ProofContext.theory_of;
    in
      thy'
    end;

(* "add_dom_fun" is a function, which maps action labels to domains, as 
 * specified in the ism definition. Warning: You have to take care yourself
 * that the mapping is unique and self-consistent.
 *)
fun add_dom_fun prep_rule types dom_free_Var transs thy =
    let
      val sign = thy;
      val (_, _, _, _, _, _, _, _, labels_Typ_default_opt, domain_Typ_default_opt) = types;

      val unit = HOLogic.unitT;
      val label_Typ = case labels_Typ_default_opt of SOME (t,_) => t
                                                   | NONE       => unit;
      val   dom_Typ = case domain_Typ_default_opt of SOME (t,_) => t
                                                   | NONE       => unit;
      val bottom_Term = Const("HOL.undefined", dom_Typ);

      (* filter label and dom out of the transition *)
      fun label_dom_from_transs prep_rule rule =
      let
        val (data_st_var_opt, rule') = transform_rule prep_rule sign types rule;

        (* expand transition *)
        val ((name, _), (_, conds)) = rule';
        val (label_opt,( dom_opt, (_, (_, (_, (_, _)))))) = conds;

        val label = case label_opt of SOME t => t | NONE =>
                     (case labels_Typ_default_opt of
                        NONE => HOLogic.unit
                      | SOME (_, SOME default_label) => default_label
                      | _ => ism_error ("ISM has labels without default labels " ^
                             "but subsection \"label\" is missing in rule " ^ quote name));
        val dom = case dom_opt of SOME t => t | NONE =>
                   (case domain_Typ_default_opt of
                      NONE => HOLogic.unit
                    | SOME (_, SOME default_dom) => default_dom
                    | _ => ism_error ("ISM has domains without default domain " ^
                           "but subsection \"dom\" is missing in rule " ^ quote name));
      in (label, dom) end;


      val vars = map (label_dom_from_transs prep_rule) transs;

      val elem_Term = Const("HOL.If", HOLogic.boolT --> dom_Typ --> dom_Typ --> dom_Typ);
      val eq_Term = Const("op =", label_Typ --> label_Typ --> HOLogic.boolT);
      fun construct_if ((label, dom), elsef) =
        let
          (* the label may contain free variables; we quantify them existentially here *)
          val free_vars = Term.add_frees label [];
          val dom_free_name = #1 (dest_Free dom_free_Var);
          val _ = if (exists ((curry (op =))(dom_free_name)) (map #1 free_vars))
            then ism_error("Sorry, but you can't use the variable name "^quote dom_free_name^" in an ISM.")
            else "";
          val dom_eq_Term = eq_Term $ dom_free_Var $ label;
          fun mk_exists2 ((name, typ), term) = HOLogic.mk_exists(name,typ,term);
          val label2 = foldl mk_exists2 dom_eq_Term free_vars;
        in
          elem_Term $ label2 $ dom $ elsef
        end;
      val if_Term = foldr construct_if bottom_Term vars;
    in
      if_Term
    end;

end;



(*** add_ism ***)

fun gen_add_ism prep_type prep_term prep_rule name params' pt_type
         pt_set_I pt_set_O msg_type cmd_type_default_opt labels_type_default_opt
         domain_type_default_opt g_st_type_opt ctrl_st data_st transs thy =
let
  val _ = message ("Constructing ism " ^ quote name ^  " ...");
  val sign = thy;

  (* open own namespace *)
  val thy = Sign.add_path name thy;
  val sign = thy;

  val params = map (fn (n,ty) => (n, mk_Typ prep_type sign ty)) params';

  (* port types and terms *)
  val pt_Typ = mk_Typ prep_type sign pt_type;
  val pt_set_Typ = HOLogic.mk_setT pt_Typ;
  val pt_set_I_Term = mk_Term prep_term sign (pt_set_I, pt_set_Typ);
  val pt_set_O_Term = mk_Term prep_term sign (pt_set_O, pt_set_Typ);

  (* message type *)
  val msg_Typ = mk_Typ prep_type sign msg_type;

  (* command type *)
  val cmd_Typ_default_opt = option_map (fn (cmd_type, default) =>
                              let val cmd_Typ = mk_Typ prep_type sign cmd_type
                              in (cmd_Typ,
                                  option_map (fn t =>
                                  mk_Term prep_term sign (t, cmd_Typ)) default)
                              end) cmd_type_default_opt;
  val cmd_Typ = case cmd_Typ_default_opt of NONE       => HOLogic.unitT
                                          | SOME (t,_) => t;

  (* label type *)
  val labels_Typ_default_opt = option_map (fn (label_type, default) =>
                              let val label_Typ = mk_Typ prep_type sign label_type
                                  val default_Term = option_map (fn t =>
                                    mk_Term prep_term sign (t, label_Typ)) default
                              in (label_Typ, default_Term)
                              end) labels_type_default_opt;
  val label_Typ = case labels_Typ_default_opt of NONE         => HOLogic.unitT
                                               | SOME (t,_) => t;

  (* domain type *)
  val domain_Typ_default_opt = option_map (fn (dom_type, default) =>
                               let val dom_Typ = mk_Typ prep_type sign dom_type
                                   val default_Term = option_map (fn t =>
                                     mk_Term prep_term sign (t, dom_Typ)) default
                               in (dom_Typ, default_Term)
                               end) domain_type_default_opt;
  val dom_Typ = case domain_Typ_default_opt of NONE         => HOLogic.unitT
                                             | SOME (t,_) => t;

  (* states *)
  (* global state type *)
  val g_st_Typ_opt = option_map (mk_Typ prep_type sign) g_st_type_opt;
  (* local state initial terms; type*)
  val ctrl_st_Typ_Term_opt = mk_state_opt prep_type prep_term sign ctrl_st;
  val data_st_Typ_Term_opt = mk_state_opt prep_type prep_term sign data_st;
  val l_st_Typ =
    case ctrl_st_Typ_Term_opt of NONE =>
          (case data_st_Typ_Term_opt of NONE =>
                let val _ = ism_error (no_state_error name) in dummyT end
            | SOME(_, data_st_Typ, _) => data_st_Typ)
    | SOME(_, ctrl_st_Typ, ctrl_st_var_dummy) =>
      case data_st_Typ_Term_opt of NONE => ctrl_st_Typ
        | SOME(_, data_st_Typ, data_st_var_dummy) =>
                 (ctrl_st_var_dummy = data_st_var_dummy; (* forcing equal type*)
                  HOLogic.mk_prodT(ctrl_st_Typ, data_st_Typ));

  (* actual state type *)
  val st_Typ = case g_st_Typ_opt of NONE           => l_st_Typ
                                  | SOME(g_st_Typ) => g_st_Typ;
  val st_set_Typ = HOLogic.mk_setT st_Typ;

  (* ism type *)
  (* the record definition itself is in ISM.thy *)
  val ism_name = "ism";
  val ism_typ_name = "ISM." ^ ism_name;
  val ism_Typ = Type (ism_typ_name, [cmd_Typ, pt_Typ, msg_Typ, st_Typ, label_Typ]);
  val param_ism_Typ = Library.foldr (op -->) (map snd params, ism_Typ);

  (* transitions *)
  val trans_Typ = Type (Sign.intern_type sign "trans",
                        [cmd_Typ, pt_Typ, msg_Typ, st_Typ, label_Typ]);
  val transs_Typ = HOLogic.mk_setT trans_Typ;
  val param_transs_Typ = Library.foldr (op -->) (map snd params, transs_Typ);
  val transs_name = "transs";

  (* domain function *)
  (* FIXME: This mapping is not necessarily unique; we should check for this and only
   * allow well-formed ISMs; it's also not correct if the label includes a parameter
   * from the state in which the action is executed!
   *)
  val dom_fun_Typ = label_Typ --> dom_Typ;
  val dom_fun_name = "dom";

  (* Small note: param_ism_Typ and param_transs_Typ handle the case that the
   * ISM is parameterised (something like 'ism AP (N::nat) =\<dots>'); in the standard
   * case, they don't change anything.
   *)

  (* add dom and ism and some other constants *)
  val thy = thy
    |> Sign.add_consts_i [(Binding.name dom_fun_name, dom_fun_Typ,       NoSyn)]
    |> Sign.add_consts_i [(Binding.name ism_name,     param_ism_Typ,     NoSyn)]
  val sign = thy

  val types = (cmd_Typ_default_opt, pt_Typ, msg_Typ, st_Typ,
               ctrl_st_Typ_Term_opt, data_st_Typ_Term_opt, l_st_Typ,
               g_st_Typ_opt, labels_Typ_default_opt, domain_Typ_default_opt);

  (* add transs (inductive definition) *)
  val transs_Term = Free (transs_name, param_transs_Typ);
  val transs_Term' = Const (Sign.full_bname thy transs_name, param_transs_Typ);
  val param_transs_Term = Library.foldl (op $) (transs_Term, map Free params);
  val param_transs_Term' = Library.foldl (op $) (transs_Term', map Free params);
  val thy = add_transs prep_rule types param_transs_Term transs thy;

  val sign = thy;
  val dom_free_Var = Free("label_var", label_Typ);
  val dom_if_Term = add_dom_fun prep_rule types dom_free_Var transs thy;
  val dom_if_eq = Const("==", dom_Typ --> dom_Typ --> propT) $
    (Const (Sign.intern_const sign dom_fun_name, dom_fun_Typ) $ dom_free_Var) $ dom_if_Term;
  val (_, thy) = thy
    |> PureThy.add_defs false [((Binding.name (dom_fun_name^"_def"), dom_if_eq), [])];

(* Earlier, a special syntax for ISM transitions was declared here. However, no code
 * uses it right now and it was broken by the introduction of domains. The code is
 * still there after all the real declaration as a comment (just for reference)
 *)

  (* initialize ism record *)
  val sign = thy;
  val typname = Sign.intern_type sign;

  (* initialize init field *)
  (* If there is no initialisation return UNIV,
   * if there is only one type of state, or both have an initial value, return a set with one element (using "insert"),
   * if there are two states (control and data), but only one field is initialized,
   *   create a lambda expression containing the condition, and convert it to a set using "Collect".
   * It may be better to create the set inductively, we will have to see\<dots>
   * TODO: All this is complicated when using a global state space --> document this!
   *)
  val boolT = HOLogic.boolT;
  fun mk_global_st_term' state = mk_global_st_term l_st_Typ g_st_Typ_opt thy state;
  val collect_const = Const("Collect", (st_Typ --> boolT) --> st_set_Typ);
  val init_var = Free("init_var", st_Typ);

  val x_var = Free("x", l_st_Typ);
  val global_x_var = mk_global_st_term' x_var;
  val x_eq = Const("op =", [st_Typ, st_Typ] ---> HOLogic.boolT) $ global_x_var $ init_var;

  (* If there is a global state space, we need a more complicated lambda-expr *)
  fun lambda_expr_l typ sel state = lambda init_var
        (Const("op =", [typ, typ] ---> boolT) $
        (Const(sel, l_st_Typ --> typ) $ init_var) $ state);
  fun lambda_expr_g typ sel state = lambda init_var
        (HOLogic.mk_exists("x", l_st_Typ, (Const("op &", [boolT, boolT] ---> boolT) $
        (Const("op =", [st_Typ, st_Typ] ---> boolT) $ global_x_var $ init_var) $
        (Const("op =", [typ, typ] ---> boolT) $
        (Const(sel, l_st_Typ --> typ) $ x_var) $ state))));
  fun lambda_expr typ sel state = case g_st_Typ_opt of NONE   => lambda_expr_l typ sel state
                                                     | SOME _ => lambda_expr_g typ sel state;

  val global_univ_term = case g_st_Typ_opt of NONE   => Const("UNIV", st_set_Typ)
                                            | SOME _ => collect_const $ (lambda init_var (HOLogic.mk_exists("x", l_st_Typ, x_eq)));
  fun one_state_init s = case s of NONE    => global_univ_term
                                 | SOME st => Const("insert", st_Typ --> st_set_Typ --> st_set_Typ) $ mk_global_st_term' st $ Const("Set.empty", st_set_Typ);

  fun two_states_init data_st_Term data_st_Typ ctrl_st_Term ctrl_st_Typ =
    case ctrl_st_Term of NONE =>
      (case data_st_Term of NONE     => global_univ_term
                          | SOME dst => collect_const $ lambda_expr data_st_Typ "snd" dst)
    | SOME cst =>
      (case data_st_Term of NONE     => collect_const $ lambda_expr ctrl_st_Typ "fst" cst
                          | SOME dst => let
                               val (pair_Const, _) = mk_pair_Const sign ctrl_st_Typ data_st_Typ;
                               val st = pair_Const $ cst $ dst
                            in
                               one_state_init (SOME st)
                            end);
  val init_Term' =
    case ctrl_st_Typ_Term_opt of NONE        =>
        (case data_st_Typ_Term_opt of NONE   => dummy_Term
        | SOME(data_st_Term, _, _)           => one_state_init data_st_Term)
    | SOME(ctrl_st_Term, ctrl_st_Typ, _)     =>
        case data_st_Typ_Term_opt of NONE    => one_state_init ctrl_st_Term
        | SOME(data_st_Term, data_st_Typ, _) =>
            two_states_init data_st_Term data_st_Typ ctrl_st_Term ctrl_st_Typ;


  (* actually define the ism using meta-equality *)
  val ism_Term = Library.foldl (op $) (Const (Sign.intern_const sign ism_name, param_ism_Typ),
                               map Free params);
  val ism_make_Term = ((Const((ism_typ_name^".make"),
    pt_set_Typ --> pt_set_Typ --> st_set_Typ --> transs_Typ --> ism_Typ)) $
    pt_set_I_Term $ pt_set_O_Term $ init_Term' $ param_transs_Term');
  val ism_eq = ((Const("==", ism_Typ --> ism_Typ --> propT)) $ ism_Term $ ism_make_Term );

  val (_, thy) = thy
    |> PureThy.add_defs false [((Binding.name "ism_def", ism_eq ), [])];

  (* close own namespace *)
  val thy = Sign.parent_path thy;
in
  thy
end;


fun assoc_string ([], (key:string)) = NONE
  | assoc_string ((keyi, xi) :: pairs, key) =
      if key = keyi then SOME xi else assoc_string (pairs, key);

fun read_rule thy sTs =
  let
    val ctxt = ProofContext.init thy;
    val ts = map (Syntax.parse_term ctxt o fst) sTs;
    val ts' = map2 (TypeInfer.constrain o snd) sTs ts;
  in map (Thm.cterm_of thy) (Syntax.check_terms ctxt ts') end;

fun cterm_of' sign = cterm_of sign o fst;
fun rule_of sign = map (cterm_of' sign);

val add_ism   = gen_add_ism read_ctyp read_cterm read_rule;
val add_ism_i = gen_add_ism ctyp_of   cterm_of'  rule_of;



(*** package setup ***)

(** outer syntax **)

local structure P = OuterParse and K = OuterKeyword in

val ism_decl =
  let val optional_init = Scan.option (P.$$$ "init" |-- P.term)
      val param = P.$$$ "(" |-- P.name -- (P.$$$ "::" |-- P.typ) --| P.$$$ ")"
      val pattern = (   (                  P.term >> (fn x => (x,false)))
                     || (P.$$$ "multi" |-- P.term >> (fn x => (x,true ))))
                    -- P.term
      val in_pat = (P.term >> (fn x => (x,false))) -- P.term
      val state_update = P.list1 (P.name -- (P.$$$ ":=" |-- P.term)) >> Inl
                         ||                                 P.term   >> Inr
  in (P.name -- Scan.repeat param) -- (P.$$$ "=" |--
    (P.$$$ "ports" |-- P.typ --
      (P.$$$ "inputs"  |-- P.term --
      (P.$$$ "outputs" |-- P.term --
    (P.$$$ "messages" |-- P.typ --
    (Scan.option (P.$$$ "commands" |-- P.typ --
                  Scan.option (P.$$$ "default" |-- P.typ)) --
    (Scan.option (P.$$$ "labels"   |-- P.typ --
                  Scan.option (P.$$$ "default" |-- P.typ)) --
    (Scan.option (P.$$$ "domains"  |-- P.typ --
                  Scan.option (P.$$$ "default" |-- P.typ)) --
    (P.$$$ "states" |-- Scan.option P.typ --
      (Scan.option (P.$$$ "control" |-- P.typ --
              (optional_init >> (fn x => (x, "dummy_control_var")))) --
      (Scan.option (P.$$$ "data"    |-- P.typ --
              (optional_init --
                (Scan.optional (P.$$$ "name" |-- P.term) "s"))) --
    Scan.optional (P.$$$ "transitions" |--
      Scan.repeat1 ((SpecParse.thm_name ":" >> apfst Binding.name_of) --
        (Scan.option (P.term -- ((P.$$$ "\<rightarrow>" || P.$$$ "->") |--
                      P.term)) --
        (Scan.option   (P.$$$ "label" |--          P.term)     --
        (Scan.option   (P.$$$ "dom"   |--          P.term)     --
        (Scan.optional (P.$$$ "pre"   |--  P.list1 P.term ) [] --
        (Scan.optional (P.$$$ "in"    |--  P.list1 in_pat ) [] --
        (Scan.optional (P.$$$ "out"   |--  P.list1 pattern) [] --
        (Scan.option   (P.$$$ "cmd"   |--          P.term)     --
        (Scan.optional (P.$$$ "post"  |--  state_update) (Inl []))
       ))))))))) []
    ))))))))))) end;

val _ =
  OuterSyntax.command "ism" "define Interacting State Machine" K.thy_decl
   (ism_decl >> (fn ((name, params), (pt_type, (pt_set_I, (pt_set_O, (msg_type,
   (cmd_type_default_opt, (labels_type_default_opt, (domains_type_default_opt, 
   (g_st_type_opt, (ctrl_st, (data_st, transs))))))))))) =>
      Toplevel.theory
        (add_ism name params pt_type pt_set_I pt_set_O msg_type
                   cmd_type_default_opt labels_type_default_opt
                   domains_type_default_opt g_st_type_opt ctrl_st
                   data_st transs)));

val _ = List.app OuterKeyword.keyword
["ports", "inputs", "outputs", "messages", "commands", "default",
 "labels", "domains", "states", "control", "init", "data", "name", "multi",
 "transitions", "\<rightarrow>", "->", "pre", "in", "out", "cmd", "label", "dom", "post",":="];

end; (* local *)


end;
*}
end
(*>*)






(*<*)
(* The syntax given below is not up-to-date any more, since the domain parameter has been introduced;
 * the code is for backward refercence only
 *)
(*
  val _ = message ("Add syntax " ^ quote name ^  " ...");
*)  (* add syntax (symbols) and translations *)
(*  val sign = thy;
  val syn_name = name ^ "_" ^ transs_name ^ "_syntax";
  val msgss_Typ = pt_Typ --> mk_listT msg_Typ;
  val symbols_Typ = Library.foldr (op -->) (map snd params,
                       msgss_Typ --> msgss_Typ --> msgss_Typ -->
                       st_Typ --> st_Typ --> HOLogic.boolT);
  val mixfix_String = (escape_underscores name) ^ Library.foldr (op ^)
    (map (fn _ => ":_") params,
    ":\<langle>_,_\<rangle> \<midarrow>_\<rightarrow>" ^
    " \<langle>_,_\<rangle>");
  val mixfix = Mixfix (mixfix_String, Library.foldr (op ::)
                   (map (fn _ => 10) params, [10, 10, 10, 10, 10]), 60);
  val param_vars = map (Variable o fst) params;
  fun mk_appl cname = Syntax.mk_appl (Constant cname);
  val trrule = mk_appl syn_name (param_vars @
                                 map Variable ["p", "s", "c", "p'", "s'"]);
  datatype name_variant = Short | Medium | Long;
  fun nv n Short  =              NameSpace.base n
  |   nv n Medium = name ^ "." ^ NameSpace.base n
  |   nv n Long  = n;
  fun state_constructor s v = case g_st_Typ_opt of NONE => s
    | SOME g_st_Typ => mk_appl (nv (fetch_constr l_st_Typ g_st_Typ thy) v) [s];
  fun tuple_AST v = mk_appl "_tuple"
      [mk_appl "_tuple"      [                   Variable "p",
       mk_appl "_tuple_arg"  [state_constructor (Variable "s") v]],
       mk_appl "_tuple_args" [                   Variable "c",
       mk_appl "_tuple_args" [                   Variable "p'",
       mk_appl "_tuple_arg"  [state_constructor (Variable "s'") v]]]];
  fun trrule' (v1, v2) = mk_appl "op :" [tuple_AST v1,
            mk_appl (nv (Sign.intern_const sign transs_name) v2) param_vars];
  val thy = thy
    |> Theory.add_modesyntax_i ("symbols", true)
                               [(syn_name, symbols_Typ, mixfix)]
    |> Theory.add_trrules_i
        (Syntax.ParsePrintRule (trrule, trrule' (Long, Long))::
*)         (* additionally for robustness wrt short/long constant names: *)
(*         map (fn vv => Syntax.PrintRule (trrule, trrule' vv))
             [(Long, Medium), (Long, Short), (Medium, Long), (Medium, Medium),
              (Medium, Short), (Short, Long), (Short, Medium), (Short, Short)]);
*)
(*>*)
